@@ -1,6 +1,10 @@
 /*
  *  Copyright (C) 2001-2003 Hewlett-Packard Co.
  *	Contributed by Stephane Eranian <eranian@hpl.hp.com>
+ *  Copyright (C) 2006-2009 Intel Corporation
+ *	Contributed by Fenghua Yu <fenghua.yu@intel.com>
+ *	Contributed by Bibo Mao <bibo.mao@intel.com>
+ *	Contributed by Chandramouli Narayanan <mouli@linux.intel.com>
  *
  * This file is part of the ELILO, the EFI Linux boot loader.
  *
@@ -30,6 +34,8 @@
 #include "ia64/sysdeps.h"
 #elif defined CONFIG_ia32
 #include "ia32/sysdeps.h"
+#elif defined CONFIG_x86_64
+#include "x86_64/sysdeps.h"
 #endif
 
 #endif /* __ELILO_SYSDEPS_H__ */
